@@ -49,12 +49,13 @@
 #include <errno.h>
 #endif
 
+#include <crc32.h>
 #include <uudeview.h>
 #include <uuint.h>
 #include <fptools.h>
 #include <uustring.h>
 
-char * uunconc_id = "$Id: uunconc.c,v 1.9 2002/04/06 02:28:35 root Exp $";
+char * uunconc_id = "$Id$";
 
 /* for braindead systems */
 #ifndef SEEK_SET
@@ -541,6 +542,8 @@ UURepairData (FILE *datei, char *line, int encoding, int *bhflag)
 
   while (vflag == 0 && nflag && safety--) {
     if (nflag == 1) {		/* need next line to repair */
+      if (strlen (line) > 250)
+	break;
       ptr = line + strlen (line);
       while (ptr>line && (*(ptr-1)=='\015' || *(ptr-1)=='\012'))
 	ptr--;
@@ -896,6 +899,69 @@ UUDecodePT (FILE *datain, FILE *dataout, int *state,
   return UURET_OK;
 }
 
+/*
+ * Decode a single field using method. For the moment, this supports
+ * Base64 and Quoted Printable only, to support RFC 1522 header decoding.
+ * Quit when seeing the RFC 1522 ?= end marker.
+ */
+
+int
+UUDecodeField (char *s, char *d, int method)
+{
+  int z1, z2, z3, z4;
+  int count=0;
+
+  if (method == B64ENCODED) {
+    while ((z1 = B64xlat[ACAST(*s)]) != -1) {
+      if ((z2 = B64xlat[ACAST(*(s+1))]) == -1) break;
+      if ((z3 = B64xlat[ACAST(*(s+2))]) == -1) break;
+      if ((z4 = B64xlat[ACAST(*(s+3))]) == -1) break;
+
+      d[count++] = (z1 << 2) | (z2 >> 4);
+      d[count++] = (z2 << 4) | (z3 >> 2);
+      d[count++] = (z3 << 6) | (z4);
+
+      s+=4;
+    }
+    if (z1 != -1 && z2 != -1 && *(s+2) == '=') {
+      d[count++] = (z1 << 2) | (z2 >> 4);
+      s+=2;
+    }
+    else if (z1 != -1 && z2 != -1 && z3 != -1 && *(s+3) == '=') {
+      d[count++] = (z1 << 2) | (z2 >> 4);
+      d[count++] = (z2 << 4) | (z3 >> 2);
+      s+=3;
+    }
+  }
+  else if (method == QP_ENCODED) {
+    while (*s && (*s != '?' || *(s+1) != '=')) {
+      while (*s && *s != '=' && (*s != '?' || *(s+1) != '=')) {
+	d[count++] = *s++;
+      }
+      if (*s == '=') {
+	if (isxdigit (*(s+1)) && isxdigit (*(s+2))) {
+	  d[count]  = (isdigit (*(s+1)) ? (*(s+1)-'0') : (tolower (*(s+1))-'a'+10)) << 4;
+	  d[count] |= (isdigit (*(s+2)) ? (*(s+2)-'0') : (tolower (*(s+2))-'a'+10));
+	  count++;
+	  s+=3;
+	}
+	else if (*(s+1) == '\012' || *(s+1) == '\015') {
+	  s+=2;
+	}
+	else {
+	  d[count++] = *s++;
+	}
+      }
+    }
+  }
+  else {
+    return -1;
+  }
+
+  d[count] = '\0';
+  return count;
+}
+
 int
 UUDecodePart (FILE *datain, FILE *dataout, int *state,
 	      long maxpos, int method, int flags,
@@ -905,11 +971,15 @@ UUDecodePart (FILE *datain, FILE *dataout, int *state,
   int warning=0, vlc=0, lc[2], hadct=0;
   int tc=0, tf=0, vflag, haddata=0, haddh=0;
   long yefilesize=0, yepartends=0;
+  crc32_t yepartcrc=crc32(0L, Z_NULL, 0);
+  static crc32_t yefilecrc=0;
   static int bhflag=0;
   size_t count=0;
+  size_t yepartsize=0;
   char *ptr;
 
   if (datain == NULL || dataout == NULL) {
+    yefilecrc = crc32(0L, Z_NULL, 0);
     bhflag = 0;
     return UURET_OK;
   }
@@ -1038,12 +1108,16 @@ UUDecodePart (FILE *datain, FILE *dataout, int *state,
       }
       else if (method == YENC_ENCODED &&
 	       strncmp (line, "=ybegin ", 8) == 0 &&
-	       _FP_strstr (line, " size=") != NULL &&
 	       _FP_strstr (line, " name=") != NULL) {
 	*state = DATA;
 
-	ptr = _FP_strstr (line, " size=") + 6;
-	yefilesize = atoi (ptr);
+	if ((ptr = _FP_strstr (line, " size=")) != NULL) {
+	  ptr += 6;
+	  yefilesize = atoi (ptr);
+	}
+	else {
+	  yefilesize = -1;
+	}
 
 	if (_FP_strstr (line, " part=") != NULL) {
 	  if (_FP_fgets (line, 1200 - 5, datain) == NULL) {
@@ -1076,6 +1150,35 @@ UUDecodePart (FILE *datain, FILE *dataout, int *state,
 
     if (*state == DATA && method == YENC_ENCODED &&
 	strncmp (line, "=yend ", 6) == 0) {
+      if ((ptr = _FP_strstr (line, " pcrc32=")) != NULL) {
+	crc32_t pcrc32 = strtoul (ptr + 8, NULL, 16);
+	if (pcrc32 != yepartcrc) {
+	  UUMessage (uunconc_id, __LINE__, UUMSG_WARNING,
+		     uustring (S_PCRC_MISMATCH), progress.curfile, progress.partno);
+	}
+      }
+      if ((ptr = _FP_strstr (line, " crc32=")) != NULL)
+      {
+	crc32_t fcrc32 = strtoul (ptr + 7, NULL, 16);
+	if (fcrc32 != yefilecrc) {
+	  UUMessage (uunconc_id, __LINE__, UUMSG_WARNING,
+		     uustring (S_CRC_MISMATCH), progress.curfile);
+	}
+      }
+      if ((ptr = _FP_strstr (line, " size=")) != NULL)
+      {
+	size_t size = atol(ptr + 6);
+	if (size != yepartsize && yefilesize != -1) {
+	  if (size != yefilesize)
+	    UUMessage (uunconc_id, __LINE__, UUMSG_WARNING,
+		       uustring (S_PSIZE_MISMATCH), progress.curfile,
+		       progress.partno, yepartsize, size);
+	  else
+	    UUMessage (uunconc_id, __LINE__, UUMSG_WARNING,
+		       uustring (S_SIZE_MISMATCH), progress.curfile,
+		       yepartsize, size);
+	}
+      }
       if (yepartends == 0 || yepartends >= yefilesize) {
 	*state = DONE;
       }
@@ -1103,6 +1206,12 @@ UUDecodePart (FILE *datain, FILE *dataout, int *state,
       if (vflag == method) {
 	if (tf) {
 	  count  = UUDecodeLine (line, oline, method);
+	  if (method == YENC_ENCODED) {
+	    if (yepartends)
+	      yepartcrc = crc32(yepartcrc, oline, count);
+	    yefilecrc = crc32(yefilecrc, oline, count);
+	    yepartsize += count;
+	  }
 	  vlc++; lc[1]++;
 	}
 	else if (tc == 3) {
@@ -1333,7 +1442,7 @@ UUDecode (uulist *data)
 
   iter = data->thisfile;
   while (iter) {
-    if (part != -1 && iter->partno != part+1)
+    if (part != -1 && iter->partno != part+1 && !uu_desperate)
       break;
     else
       part = iter->partno;
@@ -1397,7 +1506,12 @@ UUDecode (uulist *data)
        data->uudet == PT_ENCODED))
     state = DONE; /* assume we're done */
 
-  fclose (dataout);
+  if (fclose (dataout)) {
+    UUMessage (uunconc_id, __LINE__, UUMSG_ERROR,
+	       uustring (S_WR_ERR_TEMP),
+	       strerror (uu_errno = errno));
+    res = UURET_IOERR;
+  }
 
   if (res != UURET_OK || (state != DONE && !uu_desperate)) {
     unlink (data->binfile);
@@ -1568,7 +1682,14 @@ UUDecode (uulist *data)
      */
 
     fclose (datain);
-    fclose (dataout);
+    if (fclose (dataout)) {
+      UUMessage (uunconc_id, __LINE__, UUMSG_ERROR,
+		 uustring (S_WR_ERR_TARGET),
+		 ntmp, strerror (uu_errno = errno));
+      unlink (ntmp);
+      free   (ntmp);
+      return UURET_IOERR;
+    }
 
     if (unlink (data->binfile)) {
       UUMessage (uunconc_id, __LINE__, UUMSG_WARNING,
